# -*- coding: utf-8 -*-
''' Tools for sending commands to the microcontroller, as well as using the DIO '''
from time import sleep
import re
import threading

import pykarbon.hardware as pk

class Session():
    '''Attaches to terminal serial port and allows reading/writing from the port.

    Automatically performs port discovery on linux and windows. Then is able to take
    ownership of a port and perform read/write operations. Also offers a method for setting
    various mcu control properties.

    There is additional support for registering a function to certain DIO states, or input pin
    transitions. When the interface receives a registered event, it will call the function and
    optionally set the digital outputs to the returned state.
    This features requires running the session with automonitoring enabled.

    Digital IO events will be recorded in the data queues, while configuration information will
    overwite the 'info' dictionary.

    Attributes:
        interface: Serial interface object that has methods for reading/writing to the port.
        pre_data: Data before it has been parsed by the registry service.
        data: Queue for holding the data read from the port
        isopen: Bool to indicate if the interface is connected
        info: Dictionary of information about the configuration of the mcu.
        registry: Dict of registered DIO states and function responses
        bgmon: Thread object of the bus background monintor
    '''
    def __init__(self, timeout=.1, automon=True):
        '''Discovers hardware port name.

        When 'automon' is set to 'True', this object will immediately attempt to claim the terminal
        connection that it discovers. Assuming the connection can be claimed, the session will then
        start monitoring all incoming data in the background.

        This data is stored in the the session's 'data' attribute, and can be popped from the queue
        using the 'popdata' method. Additionally, the entire queue may be purged to a csv file using
        the 'storedata' method -- it is good practice to occasionally purge the queue.

        Args:
            timeout(int, optional): Time until read attempts stop in seconds. (None disables)
            automon(bool, optional): Automatically monitor incoming data in the background.
        '''
        self.interface = pk.Interface('terminal', timeout)
        self.pre_data = []
        self.data = []
        self.isopen = False

        self.info = \
        {
            'version' :
                {
                    'value': None,
                    'desc' : 'Firmware version number.'
                },
            'build' :
                {
                    'value' : None,
                    'desc' : 'Firmware build date.'
                },
            'configuration' :
                {
                    'value' : None,
                    'desc' : 'Current user configuration.'
                },
            'ignition-sense' :
                {
                    'value' : None,
                    'desc' : 'If ignition sensing is enabled or disabled.'
                },
            'startup-timer' :
                {
                    'value' : None,
                    'desc' : 'Time, in seconds, until boot after ignition on.'
                },
            'shutdown-timer' :
                {
                    'value' : None,
                    'desc' : 'Time, in seconds, until soft power off after ignition off.'
                },
            'hard-off-timer':
                {
                    'value' : None,
                    'desc' : 'Time, in seconds, until hard power off after igntion off.'
                },
            'auto-power-on':
                {
                    'value' : None,
                    'desc' : 'Force device to power on when first connected to AC power.'
                },
            'shutdown-voltage':
                {
                    'value' : None,
                    'desc' : 'Voltage when device will power off to avoid battery discharge.'
                },
            'hotplug' :
                {
                    'value' : None,
                    'desc' : 'Set if the display port outputs are hotpluggable.'
                },
            'can-baudrate' :
                {
                    'value' : None,
                    'desc' : 'Current CAN bus baudrate.'
                },
            'dio-power-switch' :
                {
                    'value' : None,
                    'desc' : 'Have digital inputs act as a remote power switch when device is off.'
                },
            'boot-config':
                {
                    'value' : None,
                    'desc' : 'If the current configuration will be loaded at boot.'
                }
        }

        self.bgmon = None
        self.registry = {}

        if automon:
            self.open()
            self.bgmonitor()
        else:
            self.data = self.pre_data

    def __enter__(self):
        if not self.isopen:
            self.interface.__enter__()
            self.isopen = True

        return self

    def open(self):
        '''Claim the interface (only one application may open the serial port)'''
        if not self.isopen:
            self.interface.claim()
            self.isopen = True

        return self.isopen

    def pushdata(self, line: str):
        '''Add data to the end of the session queue.

        NOTE: Does not push empty strings, and strips EoL characters.

        Args:
            line: Data that will be pushed onto the queue
        '''

        self.data.append(line)

    def register(self, input_num, state, action, **kwargs):
        '''Automatically perform action upon receiving data_id

        Register an action that should be automatically performed when a certain digital input
        state is read. By default, this action will only be performed when the digital input
        first transitions to a state -- subsequent bus reads will be ignored:

        Example:
            register(1, 'low', action)

            Input 1 : 1 --> 0 (Execute Action)
            Input 1 : 0 --> 0 (Do nothing)
            Input 1 : 0 --> 1 (Do nothing)
            Input 1 : 1 --> 0 (Execute Action)

        Actions should be a python function, which will be automatically wrapped in a
        pykarbon.terminal.Reactions object by this function. When the passed action is called
        Reactions will try to pass it the current dio state as the first positional argument.
        If thrown a TypeError, it will call the action without any arguments.

        There is addtional support for masking input events with a particular bus state. That
        is, if an input event occurs, but the bus does not match the state, the action will
        not be executed.

        Example:
            register(1, 'high', action, dio_state='---0 ---1')

            Input 1 : 0 --> 1, Bus State: 0011 1111 (Do nothing)
            Input 1 : 1 --> 1, Bus State: 0000 1111 (Do nothing)
            Input 1 : 1 --> 0, Bus State: 0000 1111 (Do nothing)
            Input 1 : 0 --> 1, Bus State: 0000 1111 (Execute Action)

        Note:
            Bus state format is digital output 0-4 space digital input 0-4. Dashes are 'don't care'

        Args:
            dio_state(str) : Shorthand for the state of the dio, a dash will ignore the value of .
            action: The python function that will be performed.
            kwargs:
                transition_only: Act only when a state is true by transition (Default: True)
                dio_state: Mask performing action with dio state (Default: ---- ----)
                run_in_background: Run action as background task (Default: True)
                auto_response: Automatically reply with returned message (Default: True)

        Returns:
            The 'Reaction' object that will be used in responses to this data_id
        '''

        reaction = Reactions(self.set_all_do, [input_num, state], action, **kwargs)
        self.registry[input_num] = {state : reaction}

        return reaction

    def print_info(self):
        ''' Prints out mcu configuration information '''
        top_bot = "-"
        top_bot = top_bot.rjust(37, '-')
        print(top_bot)
        for key in self.info:
            temp_val = self.info[key]['value'][0:10].ljust(12, ' ')
            temp_key = key.ljust(18, ' ')
            print("+  {}|  {}+".format(temp_key, temp_val))
        print(top_bot)

    def update_info(self, print_info=False):
        ''' Request configuration information from MCU

        Arguments:
            print(bool, optional): Print out info after update. (Default: False)
        '''

        if self.isopen:
            self.interface.cwrite('version')
            self.interface.cwrite('config')
            if print_info:
                sleep(1)
                self.print_info()

    def set_param(self, parameter: str, value: str, update=True, save_config=True):
        ''' Sets a mcu configuration parameter

        Arguments:
            parameter: Paramter to change
            value: Parameter will be set to this value
            update: Call update info to reflect param changes

        Returns:
            one or zero to indicate sucess or failure
        '''
        retvl = 0
        if self.isopen:
            self.interface.cwrite('set {} {}'.format(parameter, value))
            if update:
                sleep(.1) # Needs time to process
                self.update_info()
            if save_config:
                sleep(.1) # Needs time to process
                self.interface.cwrite('save-config')
        else:
            retvl = 1

        return retvl

    def set_do(self, number, state):
        ''' Set the state of a single digital output

        Just formats the number and state into a dictionary, and then calls the
        'set_all_do' method.

        Example:
            set_do(0, True)
            set_do('two', 0)
        '''
        set_string = []
        set_string += self.get_previous_state().split(' ')[1]
        try:
            number = int(number)
        except ValueError:
            swap = {'zero' : 0, 'one' : 1, 'two' : 2, 'three' : 3}
            number = swap[number]

        try:
            state = str(int(state))
        except ValueError:
            swap = {'low' : '0', 'high' : '1'}
            state = swap[state]

        set_string[number] = state

        self.set_all_do(set_string)
        return set_string

    def set_all_do(self, states):
        ''' Sets all digital outputs based on a list of states

        Arguments:
            states (list): A list of '1' or '0' corrospponding to the state of each output.

        Example:
            set_all_do(['0', '0', '0', '0'])  -- turn all outputs off
        '''
        self.write("set-do {}{}{}{}".format(*states))


    def parse_line(self, line):
        ''' Parse a non-dio line into mcu configuration info '''
        line = line.strip("\n\r")

        if 'err' in line.lower():
            print(line)
            return line

        found_match = True
        if '<' in line:
            version, build = line.split('|')
            self.info['version']['value'] = version.strip('<').strip(' ')
            build = build
            self.info['build']['value'] = build.strip('>').strip(' ')
        elif 'Boot' in line:
            try:
                self.info['boot-config']['value'] = line.split(':')[1].strip(' ')
            except IndexError:
                print("Unexpected response: " + line)
        elif 'Remote' in line:
            try:
                self.info['dio-power-switch']['value'] = line.split(':')[1].strip(' ')
            except IndexError:
                print("Unexpected response: " + line)
        else:
            found_match = False

        if not found_match:
            for key in self.info:
                if key in line.lower().replace(' ', '-'):
                    try:
                        self.info[key]['value'] = line.split(':')[1].strip(' ')
                    except IndexError:
                        print("Unexpected response: " + line)


        return line

    def write(self, command):
        ''' Write an arbitrary string to the serial terminal '''
        self.interface.cwrite(command)

    def readline(self):
        '''Reads a single line from the port, and stores the output in self.data

        If no data is read from the port, then nothing is added to the data queue.

        Returns
            The data read from the port
        '''
        line = ""
        if self.isopen:
            line = self.interface.cread()[0]
            dio_check = re.match(r'[0-1]{4} {0,1}[0-1]{4}', line)
            if dio_check:
                self.pre_data.append(line[0:4] + ' ' + line[4:8])
            elif line:
                self.parse_line(line)

        return line

    def bgmonitor(self):
        '''Start monitoring the terminal in the background

        Uses python threading module to start the monitoring process.

        Returns:
            The 'thread' object of this background process
        '''

        if not self.data:
            self.data = []

        self.bgmon = threading.Thread(target=self.monitor)
        self.bgmon.start()

        threading.Thread(target=self.registry_service).start()

        return self.bgmon

    def monitor(self):
        '''Watches port for incoming data while connection is open.

        The loop is predicated on the connection being open; closing the connection will stop the
        monitoring session.

        Returns:
            The method used to stop monitoring. (str)
        '''
        retvl = "SessionClosed"
        while self.isopen:
            try:
                self.readline()
            except KeyboardInterrupt:
                retvl = "UserCancelled"

        return retvl

    def registry_service(self):
        '''Check if receive line has a registered action.

        If the receive line does have an action, perform it, and then move the data
        into the main data queue. Otherwise, just move the data.
        '''
        while self.isopen:
            try:
                line = self.pre_data.pop()
                if line:
                    self.pushdata(line)
                    self.check_action(line)
            except IndexError:
                continue

        return 0

    def check_action(self, line):
        '''Check is message has an action attached, and execute if found

        Args:
            line: Dio state formatted as '[0-1]{4} [0-1]{4}'
            prev_line: The previously known state of the bus
        '''

        prev_state = self.get_previous_state(-2)

        # Determine the state of every digital input, and if it has changed state
        input_states = {}
        for i in range(1, 5):
            input_state = line[i + 4]
            prev_state = prev_state[i + 4]

            transition = False
            if input_state not in prev_state:
                transition = True

            if '0' in input_state:
                input_states[i] = {'state' : 'low', 'trans' : transition}
            else:
                input_states[i] = {'state' : 'high', 'trans' : transition}

        # Check registry against current state of each digital input
        for input_num in self.registry:
            reaction = self.registry[input_num][input_states[input_num]['state']]

            if reaction.transition_only and not input_states[input_num]['trans']:
                return

            if not re.match(reaction.dio_state.replace('-', '.'), line):
                return

            if reaction.run_in_background:
                reaction.bgstart(line)
            else:
                reaction.start(line)

        return

    def get_previous_state(self, index=-1):
        ''' Returns the previous state of the digital io '''
        try:
            prev_line = self.data[index]
        except IndexError:
            prev_line = '0000 0000'

        return prev_line


    def storedata(self, filename: str, mode='a+'):
        '''Pops the entire queue and saves it to a csv.

        This method clears the entire queue: once you have called it, all previously received
        data will no longer be stored in the sessions 'data' attribute. Instead, this data will
        now reside in the specified .csv file.

        Each received dio event has its own line of the format: outputs,inputs.

        By default, if a file that already exists is specified, the data will append to the end of
        this file. This behavior can be changed by setting 'mode' to any standard 'file.write' mode.

        Args:
            filename: Name of file that will be created.
            mode(str, optional): The file write mode to be used.
        '''

        if '.csv' not in filename:
            filename = filename + '.csv'

        with open(filename, mode) as datafile:
            while True:
                line = self.popdata()

                if not line:
                    break

                line = line.strip('\n\r')

                line = line.replace(' ', ',')
                datafile.write(line + "\n")

    def popdata(self):
        '''If there is data in the queue, pop an entry and return it.

        Uses queue behavior, so data is returned with 'first in first out' logic

        Returns:
            String of the data read from the port. Returns empty string if the queue is empty
        '''
        try:
            out = self.data.pop(0)
        except IndexError:
            out = ""

        return out

    def close(self):
        '''Release the interface so that other session may interact with it

        Any existing background monitor session will also be closed. If this session re-opens the
        connection, background monitoring will need to be manually restarted with the 'bgmonitor'
        method.
        '''
        self.isopen = False

        try:
            if self.bgmon.isAlive():
                sleep(.1)
        except AttributeError:
            sleep(.001)

        self.interface.release()

    def __exit__(self, etype, evalue, etraceback):
        self.isopen = False

        try:
            if self.bgmon.isAlive():
                sleep(.1)
        except AttributeError:
            sleep(.001)

        self.interface.__exit__(etype, evalue, etraceback)

    def __del__(self):
        if self.isopen:
            self.close()

class Reactions():
    '''A class for performing automated responses to certain dio transitions.

    If the action returns a list digital output states, then the reaction
    will set each of these states. If the action returns None, no digital
    outputs will be set.

    Example:
        Example action response: ['0', '1', '1', '0']

    Attributes:
        info: The input number and state that trigger this reaction
        dio_state: Mask reaction to this dio state
        action: Function called by this reaction
        transition_only: If the reaction will respond to non-transition events
        run_in_background: If reaction will run as background thread
        auto_response: If reaction will automatically reply
        set_do: Helper to set digital output state
    '''
    def __init__(self, set_do, info, action, **kwargs):
        '''Init attributes

        Additonally sets all kwargs to default values if they are not
        explicitly specified.
        '''
        self.set_do = set_do
        self.info = info
        self.action = action

        if 'dio_state' in kwargs:
            self.dio_state = kwargs['dio_state']
        else:
            self.dio_state = '---- ----'

        if 'transition_only' in kwargs:
            self.remote_only = kwargs['transition_only']
        else:
            self.remote_only = False

        if 'run_in_background' in kwargs:
            self.run_in_background = kwargs['run_in_background']
        else:
            self.run_in_background = True

        if 'auto_response' in kwargs:
            self.auto_response = kwargs['auto_response']
        else:
            self.auto_response = True

    def start(self, current_state):
        '''Run the action in a blocking manner

        Args:
            current_state: The current state of the dio
        '''

        try:
            out = self.action(current_state)
        except TypeError:
            out = self.action()

        return self.respond(out)

    def bgstart(self, current_state):
        '''Call start as a background thread

        Returns:
            The thread of the background action
        '''

        bgaction = threading.Thread(target=self.start, args=[current_state])
        bgaction.start()

        return bgaction

    def respond(self, returned_data):
        '''Automatically respond to frames, if requested

        Args:
            returned_data: A list of DIO states. If none, no states will be set
        '''
        if (not returned_data) or (not self.auto_response):
            return

        try:
            self.set_do(returned_data)
        except (TypeError, KeyError) as bad_return:
            print("Bad action response: ", bad_return)
            return

        return

def hardware_reference(device='K300'):
    '''Print useful hardware information about the device

    Displays hardware information about the DIO device, such as pinouts.
    Then pinouts assume that the user is facing the front if the device, and that the fins
    are pointed up.

    Args:
        device (str, optional): The karbon series being used. Defaults to the K300
    '''

    ref_k300 = \
    '''
    Info: Isolated digital input/ouput. To function properly, dio should be connect to
    external power and grund.

    Pinout: || GND | DO_1 | DO_2 | DO_3 | DO_4 | DI_1 | DI_2 | DI_3 | DI_4 | PWR ||
    '''

    ref_dict = {'K300': ref_k300}

    try:
        print(ref_dict[device.upper()])
    except KeyError:
        print("Please select from: [K300]")



''' Test pykarbon terminal functionality '''
from time import sleep
import pykarbon.terminal as pkt
STANDARD_DELAY = .75

def test_param_set():
    ''' Tests if parameters can be written to the MCU '''
    defaults = [
        ('ignition-sense', 'off'),
        #('startup-timer', '1'),
        #('shutdown-timer', '10'),
        #('hard-off-timer', '120'),
        ('auto-power-on', 'off'),
        #('shutdown-voltage', '6'),
        ('hotplug', 'on'),
        #('can-baudrate', '800'),
        ('dio-power-switch', 'off'),
        #('boot-config', 'true'),
    ]
    test_values = [
        ('ignition-sense', 'on'),
        #('startup-timer', '2'),
        #('shutdown-timer', '12'),
        #('hard-off-timer', '100'),
        ('auto-power-on', 'on'),
        #('shutdown-voltage', '7'),
        ('hotplug', 'off'),
        #('can-baudrate', '850'),
        ('dio-power-switch', 'off'),
        #('boot-config', 'false')
    ]
    out = ''
    with pkt.Session() as dev:
        for param, value in test_values:
            dev.set_param(param, value, update=False, save_config=False)
            sleep(STANDARD_DELAY)

        sleep(STANDARD_DELAY)
        dev.update_info()
        sleep(STANDARD_DELAY)

        out = dev.info

        for param, value in defaults:
            dev.set_param(param, value, update=False, save_config=False)
            sleep(STANDARD_DELAY)

    for param, value in test_values:
        assert value in out[param]['value']



# -*- coding: utf-8 -*-
'''
Pykarbon is designed to wrap interfacing with the microntroller on the Karbon series in an object,
with the explicit goal of being able to treat this object in a more pythonic manner. In this manner
most serial interactions with carbon are greatly simplified, and are more versatile.
'''

import pykarbon.can as pkc
import pykarbon.terminal as pkt

class Karbon:
    '''Handles interactions with both virtual serial ports.

    When initialized, this will scan the systems COM ports for the expected hardware/product id.
    The ports reporting this id are then attached to the class for easy recall and access.

    Attributes:
        can: A pykarbon.can session object -- used to interface with karbon can bus
        terminal: a pykarbon.terminal session object -- used to interface with karbon terminal.
    '''

    def __init__(self, automon=True):
        '''Opens a session with the two MCU virtual serial ports

        Arguments:
            automon: Defaults to True -- will cause can and terminal ports to auto-monitored

        '''
        self.can = pkc.Session(automon=automon)
        self.can.autobaud = self.autobaud #Need to override, as we have lock on terminal

        self.terminal = pkt.Session(automon=automon)

    def __enter__(self):
        self.can = self.can.__enter__()
        self.terminal = self.terminal.__enter__()

        return self

    def open(self):
        ''' Opens both ports: Only needs to be called if 'close' has been called '''
        self.can.open()
        self.terminal.open()

    def write(self, *args):
        '''Takes a command input and interperets it into a serial task:

        If given two integer args, it will send a CAN message. If given two string args, it
        will set a terminal parameter. If given an integer as the first arg, and a str as the
        second arg, it will attempt to set the corrospponding digital output.

        A single string argument will simply be sent to the terminal.

        Returns:
            None
        '''
        if len(args) == 1 and isinstance(args[0], str):
            self.terminal.write(args[0])
        elif len(args) == 2:
            if isinstance(args[0], int) and isinstance(args[1], int):
                self.can.write(args[0], args[1])
            elif isinstance(args[0], int) and isinstance(args[1], str):
                self.terminal.set_do(args[0], args[1])
            elif isinstance(args[0], str) and isinstance(args[1], str):
                self.terminal.set_param(args[0], args[1])

    def read(self, port_name='terminal', print_output=False):
        '''Get the next line sent from a port

        Args:
            port_name (str, optional): Will read from CAN if 'can' is in the port name.
                Reads from the terminal port by default.
            print_output (bool, optional): Set to false to not print read line

        Returns:
            Raw string of the line read from the port
        '''
        if 'terminal' in port_name.lower():
            line = self.terminal.readline()
        elif 'can' in port_name.lower():
            line = self.can.readline()

        if print_output:
            print(line)

        return line

    def autobaud(self, baudrate: int) -> str:
        '''Autodetect the bus baudrate

        If the passed argument 'baudrate' is None, the baudrate will be autodetected,
        otherwise, the bus baudrate will be set to the passed value.

        Args:
            baudrate: The baudrate of the bus in thousands. Set to 'None' to autodetect

        Returns:
            The discovered or set baudrate
        '''
        set_rate = None
        if not baudrate:
            self.terminal.write('can-autobaud')
            self.terminal.update_info()
            set_rate = self.terminal.info['can-baudrate']['value']
        else:
            self.terminal.set_param('can-baudrate', str(baudrate))
            set_rate = str(baudrate)

        return set_rate

    def show_info(self):
        '''Updates and prints configuration information'''
        try:
            self.terminal.update_info(print_info=True)
        except TypeError:
            self.terminal.update_info(print_info=True)

    def close(self):
        ''' Close both ports '''
        self.terminal.close()
        self.can.close()

    def __exit__(self, etype, evalue, etraceback):
        self.terminal.__exit__(etype, evalue, etraceback)
        self.can.__exit__(etype, evalue, etraceback)

    def __del__(self):
        self.terminal.__del__()
        self.can.__del__()



''' Discovery and control of hardware interfaces. '''
from sys import platform as os_type

import io
import serial

class Hardware:
    ''' Has methods for performing various hardware tasks: includes port discovery, etc.

    Attributes:
        ports (:obj:'dict'): The two virtual serial ports enumerated by the MCU.
    '''

    def __init__(self):
        ''' Discovers the MCUs two serial ports '''
        self.ports = {}
        self.get_ports()

    def get_ports(self) -> dict:
        '''Scans system serial devices and returns the two Karbon serial interfaces

        Returns:
            A dictionary with the keys 'can' and 'terminal' assigned hardware port names.
        '''
        import serial.tools.list_ports as port_list
        all_ports = port_list.comports()

        for port, desc, hwid in sorted(all_ports):
            if "1FC9:00A3" in hwid:

                if 'win' in os_type: #Fix for windows COM ports above 10
                    self.ports[self.check_port_kind(port)] = "\\\\.\\" + port
                else:
                    self.ports[self.check_port_kind(port)] = port

                desc = desc # Remove pylint warning

        return self.ports

    @staticmethod
    def check_port_kind(port_name: str) -> str:
        '''Checks if port is used for CAN or as the terminal

        Args:
            port_name: The hardware device name.

        Returns:
            The kind of port: 'can' or 'terminal'
        '''
        retvl = 'can'

        ser = serial.Serial(port_name, 115200, xonxoff=1, timeout=.1)
        sio = io.TextIOWrapper(io.BufferedRWPair(ser, ser), newline='\r')

        sio.write('version')
        sio.flush()

        if sio.readline():
            retvl = 'terminal'

        return retvl

class Interface(Hardware):
    '''Hardware subclass interface -- controls interactions with the karbon serial interfaces.

    Attributes:
        port: The hardware name of the serial interface
        ser: A serial object connection to the port.
        sio: An io wrapper for the serial object.
        multi_line_response: The number of lines returned when special commands are transmitted.
    '''
    def __init__(self, port_name: str, timeout=.1):
        ''' Opens a connection with the terminal port

        Args:
            port_name: Human-readable name of serial port ("can" or "terminal")
        '''
        super(Interface, self).__init__()
        self.port = self.ports[port_name]
        self.ser = None
        self.sio = None
        self.multi_line_response = {"config" : 12, "status" : 5}
        self.timeout = timeout

    def claim(self):
        '''Claims the serial interface for this instance.'''
        self.ser = serial.Serial(self.port, 115200, xonxoff=1, timeout=self.timeout)
        self.sio = io.TextIOWrapper(io.BufferedRWPair(self.ser, self.ser), newline='\r')

    def __enter__(self):
        self.claim()
        return self

    def cwrite(self, command: str):
        '''Writes a command string to the serial terminal and gets the response.

        Args:
            command: Action to be executed on the mcu

        Returns:
            None
        '''

        try:
            self.sio.flush() # Try to clear junk
            self.sio.write(command)
            self.sio.flush() # Write NOW
        except AttributeError as error:
            raise ConnectionError("Port may not be claimed; see 'claim' method") from error

    def cread(self, nlines=1):
        '''Reads n lines from the serial terminal.

        Args:
            nlines(int, optional): How many lines to try and read

        Returns:
            The combined output of each requested read transaction
        '''
        output = []
        try:
            for line in range(0, nlines):
                line = self.sio.readline()
                output.append(line)
        except AttributeError as error:
            raise ConnectionError("Port may not be claimed; see 'claim' method") from error

        return output

    def release(self):
        '''Release the interface, and allow other applications to use this port'''
        try:
            self.ser.close()
            self.sio = None
            self.ser = None
        except AttributeError:
            return

    def __exit__(self, etype, evalue, etraceback):
        self.release()
        return True

    def __del__(self):
        self.release()



''' Tests for the basic underlying tools -- discovering ports, etc. '''
import re
import pykarbon.hardware as pk

def test_port_discovery():
    ''' Test that both ports are discovered '''
    found_ports = pk.Hardware().get_ports()

    assert 'can' in found_ports
    assert 'terminal' in found_ports

def test_port_claim_release():
    ''' Confirm that ports can be claimed and released by pyserial '''
    term = pk.Interface('terminal')
    can = pk.Interface('can')

    term.claim()
    can.claim()

    assert term.ser.is_open
    assert can.ser.is_open

    term.release()
    can.release()

    assert not term.ser
    assert not can.ser

def test_read_write():
    ''' Confirm that we can write to both ports '''

    term = pk.Interface('terminal')
    term.claim()

    can = pk.Interface('can')
    can.claim()

    term.cwrite('version')
    can.cwrite('std 123 8 1122334455667788 data')

    term_out = term.cread()
    can_out = can.cread()

    match = re.match(r"<.+v(\d\.){3}\d.+>", term_out[0])
    assert match
    assert '123 1122334455667788' in can_out[0]

def test_context_manager():
    ''' Check that we can use the interface as a context manager '''

    out = ''
    with pk.Interface('terminal') as term:
        term.cwrite('version')
        out = term.cread()

    match = re.match(r"<.+v(\d\.){3}\d.+>", out[0])
    assert match

def test_multiple_release():
    ''' Ensure that is allowed to close a closed port '''
    term = pk.Interface('terminal')
    term.claim()

    term.release()
    term.release()
    term.release()

def test_unclaimed_read_write():
    ''' Check that correct errors are thrown when a user tries a closed port '''
    from pytest import raises
    term = pk.Interface('terminal')

    with raises(ConnectionError):
        term.cwrite('version')

    with raises(ConnectionError):
        term.cread()

def test_multiple_argument_writes():
    ''' Check that buffer is cleared in such a way that multi-argument writes always succeed '''
    test_values = ['700', '710', '720', '730', '740', '750', '800']
    out_values = []
    with pk.Interface('terminal') as term:
        for value in test_values:
            term.cwrite('set can-baudrate {}'.format(value))
            out_values.append(term.cread()[0])

    for value in out_values:
        assert 'Error' not in value



# -*- coding: utf-8 -*-
''' Tool for running a session with the can interface '''
from time import sleep
import threading

import pykarbon.hardware as pk

class Session():
    '''Attaches to CAN serial port and allows reading/writing from the port.

    Automatically performs port discovery on linux and windows. Then is able to take
    ownership of a port and perform read/write operations. Also offers an intelligent
    method of sending can messages that will automatically determine frame format, type,
    and data length based only on the message id and data.

    There is additional support for registering a function to certain can data ids. When the
    interface receives a registered message, it will call the function and send the returned
    data. This features requires running the session with automonitoring enabled.

    By default, the session will also try to automatically discover the bus baudrate.

    Attributes:
        interface: Serial interface object that has methods for reading/writing to the port.
        pre_data: Data before it has been parsed by the registry service.
        data: Queue for holding the data read from the port
        isopen: Bool to indicate if the interface is connected
        baudrate: Reports the discovered or set baudrate
        registry: Dict of registered DIO states and function responses
        bgmon: Thread object of the bus background monintor
    '''
    def __init__(self, baudrate=None, timeout=.1, automon=True):
        '''Discovers hardware port name.

        If the baudrate option is left blank, the device will instead attempt to automatically
        detect the baudrate of the can-bus. When 'automon' is set to 'True', this object will
        immediately attempt to claim the CAN connection that it discovers. Assuming the connection
        can be claimed, the session will then start monitoring all incoming data in the background.

        This data is stored in the the session's 'data' attribute, and can be popped from the queue
        using the 'popdata' method. Additionally, the entire queue may be purged to a csv file using
        the 'storedata' method -- it is good practice to occasionally purge the queue.

        Args:
            baudrate(int, optional): Specify a baudrate, in thousands (500 -> 500K).
            timeout(int, optional): Time until read/write attempts stop in seconds. (None disables)
            automon(bool, optional): Automatically monitor incoming data in the background.
        '''
        self.interface = pk.Interface('can', timeout)
        self.pre_data = []
        self.data = []
        self.isopen = False
        self.baudrate = self.autobaud(baudrate)
        self.bgmon = None
        self.registry = {}

        if automon:
            self.open()
            self.bgmonitor()
        else:
            self.data = self.pre_data

    def __enter__(self):
        if not self.isopen:
            self.interface.__enter__()
            self.isopen = True

        return self

    def open(self):
        '''Claim the interface (only one application may open the serial port)'''
        if not self.isopen:
            self.interface.claim()
            self.isopen = True

        return self.isopen

    def pushdata(self, line: str):
        '''Add data to the end of the session queue.

        NOTE: Strips EoL characters.

        Args:
            line: Data that will be pushed onto the queue
        '''

        self.data.append(line.strip('\n\r'))

    @staticmethod
    def autobaud(baudrate: int) -> str:
        '''Autodetect the bus baudrate

        If the passed argument 'baudrate' is None, the baudrate will be autodetected,
        otherwise, the bus baudrate will be set to the passed value.

        Args:
            baudrate: The baudrate of the bus in thousands. Set to 'None' to autodetect

        Returns:
            The discovered or set baudrate
        '''
        set_rate = None
        with pk.Interface('terminal') as term:
            if not baudrate:
                term.cwrite('can-autobaud')

                i = 0
                set_rate = term.cread()[0].strip('\n\r')
                while not set_rate and i < 2000:
                    set_rate = term.cread()[0].strip('\n\r')
                    i += 1
            else:
                term.cwrite('set can-baudrate ' + str(baudrate))
                set_rate = str(baudrate)

        return set_rate

    @staticmethod
    def format_message(data_id, data, **kwargs):
        ''' Takes an id and data and determines other message characteristics

        When keyword arguments are left blank, this function will extrapolate the correct
        frame information based on the characteristics of the passed id and data.
        If desired, all of the automatically determined characteristics may be overwritten.

        Args:
            data_id(int): Data id of the message, in hex
            data(int): Message data, in hex -- if 'None', the device will send a remote frame.
            **kwargs:
                'format': Use standard or extended frame data id ('std' or 'ext')
                'length': Length of data to be transmitted, in bytes (11223344 -> 4)
                'type': Type of frame ('remote' or 'data')
        '''

        message = {'format': 'std', 'id': data_id, 'length': 0, 'data': data, 'type': 'data'}

        if data_id > 0x7FF:
            message['format'] = 'ext'

        message['id'] = hex(data_id)[2:]

        if not data:
            message['type'] = 'remote'
        else:
            data = hex(data)[2:]
            message['length'] = int(len(data) / 2)
            message['data'] = data

        for key, value in kwargs:
            try:
                message[key] = value
            except KeyError:
                print("{} is not a valid keyword".format(key))

        return message

    def send_can(self, message) -> str:
        '''Transmits the passed message on the canbus

        Args:
            message: A dictionary containing the data required to build a can message

        Returns:
            The string version of the transmitted message
        '''

        str_message = '{format} {id} {length} {data} {type}'.format(**message)
        self.interface.cwrite(str_message)

        # Encourage io to actually send packets
        sleep(.0001)

        return str_message

    def register(self, data_id, action, **kwargs):
        '''Automatically perform action upon receiving data_id

        Register an action that should be automatically performed when a certain data
        id is read. By default the action will be performed when the id is attached
        to any frame type, and the action's returned data will be checked -- if the data
        can be formatted as a can message, it will automatically be transmitted as a reply.

        Actions should be a python function, which will be automatically wrapped in a
        pykarbon.can.Reactions object by this function. When the passed action is called
        Reactions will try to pass it the hex id and data as the first and second positional
        arguments. If thrown a TypeError, it will call the action without any arguments.

        NOTE: If the frame is a remote request frame, the passed data will be 'remote' instead
        of an int!

        Args:
            data_id(int): The hex data_id that the action will be registered to
            action: The python function that will be performed.
            kwargs:
                remote_only: Respond only to remote request frames (Default: False)
                run_in_background: Run action as background task (Default: True)
                auto_response: Automatically reply with returned message (Default: True)

        Returns:
            The 'Reaction' object that will be used in responses to this data_id
        '''

        reaction = Reactions(self.write, data_id, action, **kwargs)
        self.registry[data_id] = reaction

        return reaction

    def write(self, can_id, data):
        '''Auto-format and transmit message

        For the large majority of use cases, this is the simplest and best method to send a packet
        of data over the canbus. Only message id and the data need to specified as hex values. All
        other information about the packet will be extrapolated.

        Args:
            can_id: The hex id of the data
            data: The hex formatted data
        '''
        message = self.format_message(can_id, data)
        self.send_can(message)


    def readline(self):
        '''Reads a single line from the port, and stores the output in self.data

        If no data is read from the port, then nothing is added to the data queue.

        Returns
            The data read from the port
        '''
        line = ""
        if self.isopen:
            line = self.interface.cread()[0]
            if line:
                self.pre_data.append(line)

        return line

    def bgmonitor(self):
        '''Start monitoring the canbus in the background

        Uses python threading module to start the monitoring process.

        Returns:
            The 'thread' object of this background process
        '''

        if not self.data:
            self.data = []

        self.bgmon = threading.Thread(target=self.monitor)
        self.bgmon.start()

        threading.Thread(target=self.registry_service).start()

        return self.bgmon

    def monitor(self):
        '''Watches port for can data while connection is open.

        The loop is predicated on the connection being open; closing the connection will stop the
        monitoring session.

        Args:
            session: A canbus session object

        Returns:
            The method used to stop monitoring. (str)
        '''
        retvl = "SessionClosed"
        while self.isopen:
            try:
                self.readline()
            except KeyboardInterrupt:
                retvl = "UserCancelled"

        return retvl

    def registry_service(self):
        '''Check if receive line has a registered action.

        If the receive line does have an action, perform it, and then move the data
        into the main data queue. Otherwise, just move the data.
        '''
        while self.isopen:
            try:
                line = self.pre_data.pop(0)
                if line:
                    self.check_action(line)
                    self.pushdata(line)
            except IndexError:
                continue

        return 0

    def check_action(self, line):
        '''Check is message has an action attached, and execute if found

        Args:
            Can message formatted as [id] [data]
        '''
        try:
            data_id, message = line.strip('\n\r').split(' ')
        except ValueError:
            return

        data_id = int(data_id, 16)

        if data_id in self.registry:
            reaction = self.registry[data_id]
            if reaction.remote_only and ("remote" not in message):
                return

            if reaction.run_in_background:
                reaction.bgstart(message)
            else:
                reaction.start(message)

        return


    def storedata(self, filename: str, mode='a+'):
        '''Pops the entire queue and saves it to a csv.

        This method clears the entire queue: once you have called it, all previously received
        data will no longer be stored in the sessions 'data' attribute. Instead, this data will
        now reside in the specified .csv file.

        Each received can message has its own line of the format: id,data.

        By default, if a file that already exists is specified, the data will append to the end of
        this file. This behavior can be changed by setting 'mode' to any standard 'file.write' mode.

        Args:
            filename: Name of file that will be created.
            mode(str, optional): The file write mode to be used.
        '''

        if '.csv' not in filename:
            filename = filename + '.csv'

        with open(filename, mode) as datafile:
            while True:
                line = self.popdata()

                if not line:
                    break

                line = line.strip('\n\r')

                line = line.replace(' ', ',')
                datafile.write(line + "\n")

    def popdata(self):
        '''If there is data in the queue, pop an entry and return it.

        Uses queue behavior, so data is returned with 'first in first out' logic

        Returns:
            String of the data read from the port. Returns empty string if the queue is empty
        '''
        try:
            out = self.data.pop(0)
        except IndexError:
            out = ""

        return out

    def close(self):
        '''Release the interface so that other session may interact with it

        Any existing background monitor session will also be closed. If this session re-opens the
        connection, background monitoring will need to be manually restarted with the 'bgmonitor'
        method.
        '''
        self.isopen = False

        try:
            if self.bgmon.isAlive():
                sleep(.1)
        except AttributeError:
            sleep(.001)

        self.interface.release()

    def __exit__(self, etype, evalue, etraceback):
        self.isopen = False

        try:
            if self.bgmon.isAlive():
                sleep(.1)
        except AttributeError:
            sleep(.001)

        self.interface.__exit__(etype, evalue, etraceback)

    def __del__(self):
        if self.isopen:
            self.close()

class Reactions():
    '''A class for performing automated responses to certain can messages.

    If the action returns a dict of hex id and data, then the reaction will
    automatically respond with this id and data. If the dict has 'None' for
    id, then the reaction will respond with the originating frame's id and
    then returned data.

    Example:
        Example action response: {'id': 0x123, 'data': 0x11223344}

    Attributes:
        data_id: The can data id registered with this reaction
        action: Function called by this reaction
        remote_only: If the reaction will respond to non-remote request frames
        run_in_background: If reaction will run as background thread
        auto_response: If reaction will automatically reply
        canwrite: Helper to write out can messages
    '''
    def __init__(self, canwrite, data_id, action, **kwargs):
        '''Init attributes

        Additonally sets all kwargs to default values if they are not
        explicitly specified.
        '''
        self.canwrite = canwrite
        self.data_id = data_id
        self.action = action

        if 'remote_only' in kwargs:
            self.remote_only = kwargs['remote_only']
        else:
            self.remote_only = False

        if 'run_in_background' in kwargs:
            self.run_in_background = kwargs['run_in_background']
        else:
            self.run_in_background = True

        if 'auto_response' in kwargs:
            self.auto_response = kwargs['auto_response']
        else:
            self.auto_response = True

    def start(self, hex_data):
        '''Run the action in a blocking manner

        Args:
            hex_data: The hex data of the message that invoked this reaction.
                Should be the string 'remote' for remote frames.
        '''
        if not self.remote_only and ('remote' not in hex_data):
            hex_data = int(hex_data, 16)

        try:
            out = self.action(self.data_id, hex_data)
        except TypeError:
            out = self.action()

        return self.respond(out)

    def bgstart(self, hex_data):
        '''Call start as a background thread

        Returns:
            The thread of the background action
        '''

        bgaction = threading.Thread(target=self.start, args=[hex_data])
        bgaction.start()

        return bgaction

    def respond(self, returned_data):
        '''Automatically respond to frames, if requested

        Args:
            returned_data: A dict of id and data. If None, no response will be sent
        '''
        if (not returned_data) or (not self.auto_response):
            return

        try:
            if not returned_data['id']:
                self.canwrite(self.data_id, returned_data['data'])
            else:
                self.canwrite(returned_data['id'], returned_data['data'])
        except (TypeError, KeyError) as bad_return:
            print("Bad action response: ", bad_return)
            return

        return

def hardware_reference(device='K300'):
    '''Print useful hardware information about the device

    Displays hardware information about the CAN device, such as pinouts.
    Then pinouts assume that the user is facing the front if the device, and that the fins
    are pointed up.

    Args:
        device (str, optional): The karbon series being used. Defaults to the K300
    '''

    ref_k300 = \
    '''
    Info: Compliant with CAN 2.0B. The canbus is not internally terminated; the device
    should be used with properly terminated CAN cables/bus. The termination resistors
    are required to match the nominal impedance of the cable. To meet ISO 11898, this
    resistance should be 120 Ohms.

    Pinout: || GND | CAN_LOW | CAN_HIGH ||
    '''

    ref_dict = {'K300': ref_k300}

    try:
        print(ref_dict[device.upper()])
    except KeyError:
        print("Please select from: [K300]")



# -*- coding: utf-8 -*-
''' Update the karbon's firmware, given a firmware file '''

import sys
from time import sleep
import pykarbon.hardware as pk

ERR_CD = \
{
    0 : 'Sucess: Firmware Updated!',
    1 : 'Error : Wrong number of arguments.',
    2 : 'Error : Please specify a binary file.',
    3 : 'Error : Could not open firmware file!',
}

def check_file(filename):
    ''' Check if update file is valid '''
    rtval = 0

    if ".bin" not in filename:
        rtval = 2

    try:
        test = open(filename, 'rb')
        test.close()
    except FileNotFoundError:
        rtval = 3

    return rtval

def update_firmware(binary_file):
    ''' Update the K300 firware with specified binary update file '''

    print("Starting update...")

    # Check file is OK
    file_error = check_file(binary_file)
    if file_error:
        return file_error

    print("Launching bootloader...")

    dev = pk.Hardware()

    # Enter the bootloader, if we are in firmware
    if 'terminal' in dev.ports:
        with pk.Interface('terminal') as term:
            term.cwrite('launch-bootloader')

        sleep(1)

    # Claim an interface
    with pk.Interface('can') as bootloader:
        # Perform flash sequence
        flash_seq = ['start', 'erase', 'flash']

        for step in flash_seq:
            print("\r\bPerforming step... " + step, end='')
            bootloader.cwrite(step)

        print("\nFlashing...", end='')
        with open(binary_file, 'rb') as update:
            bootloader.ser.write(update.read())


    print("Done!")
    return 0

if __name__ == "__main__":
    if len(sys.argv) == 2:
        print(ERR_CD[update_firmware(sys.argv[1])])
    else:
        print(ERR_CD[1])



''' Testing the can tools. Assumes the Karbon has all can messages echoed'''
from time import sleep
import pykarbon.can as pkc

STANDARD_DELAY = .75

# --------------------------------------------------------------------------------------
ACTION_ID = 0
ACTION_DATA = 0
def action_with_args(hex_id, hex_data):
    ''' Action called by CAN registry service where data is passed '''
    global ACTION_ID
    global ACTION_DATA

    ACTION_ID = hex_id
    ACTION_DATA = hex_data

def action_no_args():
    ''' Action called by CAN registry service where no data is passed '''
    return {'id' : 0x111, 'data': 0x22}

# -------------------------------------------------------------------------------------

def test_open_close():
    ''' Check the port can be opened and closed '''
    dev = pkc.Session(baudrate='800', automon=False)

    assert not dev.isopen

    dev.open()

    assert dev.isopen

    dev.close()

    assert not dev.isopen

def test_push_pop():
    ''' Confirm that data is pushed and popped from queue correctly '''
    dev = pkc.Session(baudrate='800', automon=False)

    assert not dev.data

    dev.pushdata('12 112233')
    assert len(dev.data) == 1
    assert '12 112233' in dev.data[-1]

    dev.pushdata('34 445566')
    assert len(dev.data) == 2
    assert '34 445566' in dev.data[-1]

    out = dev.popdata()
    assert len(dev.data) == 1
    assert '12 112233' in out

    out = dev.popdata()
    assert not dev.data
    assert '34 445566' in out

def test_automon_start_stop():
    ''' Check that automatic port monitoring will be correctly started and stopped '''
    dev = pkc.Session(baudrate='800')

    assert dev.isopen
    assert dev.bgmon.isAlive()

    dev.close()

    assert not dev.isopen
    assert not dev.bgmon.isAlive()

def test_automon_restart():
    ''' Check that automonitoring can be stopped and restarted '''

    dev = pkc.Session(baudrate='800')

    dev.close()

    dev.open()
    dev.bgmonitor()

    assert dev.isopen
    assert dev.bgmon.isAlive()

    dev.close()

def test_autobaud():
    ''' Check that autobaudrate correctly discovers the bus baudrate '''

    dev = pkc.Session()
    dev.close()

    assert 'S' in dev.baudrate[0]

def test_format_message():
    ''' Check that messages can be formatted into a dictionary correctly '''

    dev = pkc.Session(baudrate='800', automon=False)

    message = dev.format_message(0x123, 0x11223344)
    expected = {'format': 'std', 'id': '123', 'length': 4, 'data': '11223344', 'type': 'data'}

    assert message == expected

    message = dev.format_message(0x800, 0x1122334455)
    expected = {'format': 'ext', 'id': '800', 'length': 5, 'data': '1122334455', 'type': 'data'}

    assert message == expected

    message = dev.format_message(0x123, None)
    expected = {'format': 'std', 'id': '123', 'length': 0, 'data': None, 'type': 'remote'}

    assert message == expected

def test_read_write():
    ''' Check that can is able to read and write '''
    dev = pkc.Session(baudrate='800', automon=False)
    message = {'format': 'std', 'id': '123', 'length': 4, 'data': '11223344', 'type': 'data'}

    dev.open()

    sent = dev.send_can(message)
    assert 'std 123 4 11223344 data' in sent

    sleep(STANDARD_DELAY)
    out = dev.readline()
    assert '123 11223344' in out


def test_auto_read_write():
    ''' Confirm that reading and writing still works when automonitoring '''
    dev = pkc.Session(baudrate='800')
    dev.write(0x123, 0x1122334455667788)

    sleep(STANDARD_DELAY)

    dev.close()

    out = dev.popdata()
    assert '123 1122334455667788' in out

def test_storedata():
    ''' Test that we are able to save our data to a csv '''
    from os import remove
    dev = pkc.Session(baudrate='800', automon=False)

    line_one = '123 112233'
    line_two = '456 445566'
    dev.pushdata(line_one)
    dev.pushdata(line_two)

    dev.storedata('test', mode='w+')

    out = []
    with open('test.csv', 'r') as testfile:
        out = testfile.read().split()

    assert line_one.replace(' ', ',') in out[0]
    assert line_two.replace(' ', ',') in out[1]

    remove('test.csv')

def test_context_manager():
    ''' Test that we can use can session in a context manager '''

    with pkc.Session() as dev:
        assert dev.isopen
        assert dev.bgmon.isAlive()

        dev.write(0x123, 0x11223344)
        sleep(STANDARD_DELAY)

        out = dev.popdata()
        assert '123 11223344' in out

def test_can_register():
    ''' Test that we can register and perform actions '''

    dev = pkc.Session(baudrate='800', automon=True)

    dev.register(0x777, action_with_args)
    dev.register(0x666, action_no_args)

    assert dev.registry[0x777].data_id == 0x777
    assert dev.registry[0x666].data_id == 0x666

    dev.pre_data.append('777 11223344')
    dev.pre_data.append('666 11223344')
    sleep(STANDARD_DELAY)
    dev.close()

    assert ACTION_ID == 0x777
    assert ACTION_DATA == 0x11223344
    assert '111 22' in dev.data



'''
Setup file that configures the package.
'''
from setuptools import setup

setup(
    name='pykarbon',
    url='https://github.com/onlogic/Pykarbon',
    author='Logic Supply',
    author_email='jacob.caughfield@logicsupply.com',
    packages=['pykarbon'],
    install_requires=['pyserial'],
    scripts=['scripts/update_firmware.py'],
    version='1.0.5',
    license='BSD-2.0',
    description='Tools for Karbon hardware interfaces.',
    long_description=open('README.rst').read(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
    ]
)



''' Tests covering the integrated pykarbon module '''
from time import sleep
import pykarbon.pykarbon as pk

STANDARD_DELAY = .75

def test_close_open():
    ''' Tests that a connection can be opened and closed '''

    dev = pk.Karbon()

    assert dev.can.isopen
    assert dev.terminal.isopen

    dev.close()

    assert not dev.can.isopen
    assert not dev.can.isopen

def test_context_manager():
    ''' Tests that we can use pykarbon as a context managed system '''

    with pk.Karbon() as dev:
        assert dev.can.isopen
        assert dev.terminal.isopen

def test_can_write():
    ''' Test that we can write to the can bus '''

    out = ''
    with pk.Karbon() as dev:
        dev.write(0x123, 0x11223344)
        sleep(STANDARD_DELAY)
        out = dev.can.popdata()

    assert '123 11223344' in out

def test_do_set():
    ''' Test that we can set digital outputs '''
    out = ''
    with pk.Karbon() as dev:
        dev.write(0, '1')
        sleep(STANDARD_DELAY)
        dev.write(0, '0')
        out = dev.terminal.popdata()

    assert out

def test_param_set(capsys):
    ''' Check that we can set configuration parameters '''
    out = ''
    with pk.Karbon() as dev:
        dev.write('can-baudrate', '750')

        sleep(STANDARD_DELAY)
        out = (dev.terminal.info['can-baudrate']['value'])
        dev.write('can-baudrate', '800')

    captured = capsys.readouterr()

    assert 'Error' not in captured.out
    assert '750' in out

def test_write_generic_string():
    ''' Confirm that we can write a generic string to the port '''
    from re import match

    dev = pk.Karbon(automon=False)
    dev.open()
    dev.write('version')
    sleep(STANDARD_DELAY)
    out = dev.read()

    dev.close()

    check = match(r"<.+v(\d\.){3}\d.+>", out)
    assert check

def test_show_info():
    ''' Test that we can show info '''
    with pk.Karbon() as dev:
        dev.show_info()
